"""
Obsplot jsdom handling.
"""

import subprocess
import json
import shutil
from IPython.display import HTML, SVG

from typing import Any, Optional

from .parsing import SpecParser


class ObsplotJsdom:
    """Obsplot JSDom class.

    The class takes a plot specification as input and generates a plot as SVG or HTML
    by calling a JSDom script with node.

    The specification can be given as a dict, a Plot function call or as
    Python kwargs.
    """

    def __init__(self, spec: Any) -> None:
        """
        Constructor. Parse the spec given as argument.
        """
        # Create parser
        parser = SpecParser("jsdom")
        # Parse spec code
        code = parser.parse(spec)
        # Create spec object
        spec = {"data": parser.serialize_data(), "code": code}
        self.spec = spec

    def plot(self):
        """Generates the plot by calling node script.

        Returns:
            Either an HTML or SVG IPython.display object.
        """

        # Check for node executable
        npx = shutil.which("npx")
        if not npx:
            raise RuntimeError("npx executable has not been found.")
        # Run node script with JSON spec as input
        p = subprocess.run(
            ["npx", "pyobsplot"],
            input=json.dumps(self.spec),
            capture_output=True,
            encoding="Utf8",
        )
        if p.returncode != 0:
            raise RuntimeError(
                f"jsdom script error (${p.returncode}): ${p.stderr} - ${p.stdout}"
            )
        # Get script output
        out = p.stdout
        # If output is svg, returns IPython.display.SVG
        if out[0:4] == "<svg":
            return SVG(out)
        # Else, returns IPython.display.HTML
        else:
            return HTML(out)



"""
Plot specification parsing.
"""

import datetime
import base64
import pandas as pd
import polars as pl

from typing import Any, Optional

from .data import pd_to_arrow, pl_to_arrow


class SpecParser:
    """
    Class implementing plot specification parsing.
    """

    def __init__(self, renderer: str = "widget") -> None:
        """
        SpecParser constructor.

        Args:
            renderer(str): type of renderer ("widget" or "jsdom").
        """
        self.renderer = renderer
        self.data = []

    def cache_index(self, data: Any) -> Optional[int]:
        """Returns the index of a data object in the data cache.

        Args:
            data (Any): a data object (DataFeame, GeoJson...)

        Returns:
            Optional[int]: index of the data object in the cache, or None if absent.
        """
        index = [i for i, d in enumerate(self.data) if d is data]
        if len(index) == 1:
            return index[0]
        return None

    def parse(self, spec: Any) -> Any:
        """Recursively parse a Plot specification to check and convert its elements.

        Args:
            spec (Any): a complete specification or part of a specification.

        Returns:
            Any: parsed specification or part of a specification.
        """
        if spec is None:
            return None
        # If list or tuple, recursively parse elements
        if isinstance(spec, list) or isinstance(spec, tuple):
            return [self.parse(s) for s in spec]
        # If Geojson, handle caching, don't parse, add type and returns as is
        if (
            isinstance(spec, dict)
            and "type" in spec
            and spec["type"] == "FeatureCollection"
        ):
            index = self.cache_index(spec)
            if index is None:
                self.data.append(spec)
                return {"pyobsplot-type": "GeoJson-ref", "value": (len(self.data) - 1)}
            else:
                return {"pyobsplot-type": "GeoJson-ref", "value": index}
        # If dict, parse recursively
        if isinstance(spec, dict):
            return {k: self.parse(v) for k, v in spec.items()}
        # If pandas DataFrame, handle caching, add type and serialize to Arrow IPC
        if isinstance(spec, pd.DataFrame):
            index = self.cache_index(spec)
            if index is None:
                self.data.append(spec)
                return {
                    "pyobsplot-type": "DataFrame-ref",
                    "value": (len(self.data) - 1),
                }
            else:
                return {"pyobsplot-type": "DataFrame-ref", "value": index}
        # If polars DataFrame, handle caching, add type and serialize to Arrow IPC
        if isinstance(spec, pl.DataFrame):
            index = self.cache_index(spec)
            if index is None:
                self.data.append(spec)
                return {
                    "pyobsplot-type": "DataFrame-ref",
                    "value": (len(self.data) - 1),
                }
            else:
                return {"pyobsplot-type": "DataFrame-ref", "value": index}
        # If pandas Series, convert to DataFrame and parse
        if isinstance(spec, pd.Series):
            return self.parse(pd.DataFrame(spec))
        # If polars Series, convert to DataFrame and parse
        if isinstance(spec, pl.Series):
            return self.parse(pl.DataFrame(spec))
        # If date or datetime, add tupe and convert to isoformat.
        if isinstance(spec, datetime.date) or isinstance(spec, datetime.datetime):
            return {"pyobsplot-type": "datetime", "value": spec.isoformat()}
        # Handling of JavaScript methods as objects, such as "Math.sin"
        # Manually call the parsed result ans add a special "function-object" type
        if (
            callable(spec)
            and isinstance(spec(), dict)
            and spec()["pyobsplot-type"] == "function"
        ):
            out = spec()
            out["pyobsplot-type"] = "function-object"
            return out
        return spec

    def serialize_data(self) -> list:
        """Serialize data in the data cache.

        Returns:
            list: list of serialized data objects.
        """
        result = []
        for d in self.data:
            # If polars DataFrame, serialize to Arrow IPC
            if isinstance(d, pl.DataFrame):
                value = pl_to_arrow(d)
                if self.renderer == "jsdom":
                    value = base64.standard_b64encode(value).decode("ascii")
                result.append({"pyobsplot-type": "DataFrame", "value": value})
            # If pandas DataFrame, serialize to Arrow IPC
            elif isinstance(d, pd.DataFrame):
                value = pd_to_arrow(d)
                if self.renderer == "jsdom":
                    value = base64.standard_b64encode(value).decode("ascii")
                result.append({"pyobsplot-type": "DataFrame", "value": value})
            # Else, keep as is
            else:
                result.append(d)
        return result


class JSModule(type):
    """metaclass to allow JavaScript module and methods handling."""

    def __getattr__(cls: type, name: str) -> callable:
        """Intercept methods calling and returns a parsed and typed dict object."""

        def wrapper(*args, **kwargs) -> dict:
            if kwargs:
                raise ValueError(
                    f"kwargs must not be passed to {cls.__name__}.{name} : {kwargs}"
                )
            return {
                "pyobsplot-type": "function",
                "module": cls.__name__,
                "method": name,
                "args": args,
            }

        return wrapper


class Plot(metaclass=JSModule):
    """JSModule class to allow Plot objects in specification."""

    pass


class d3(metaclass=JSModule):
    """JSModule class to allow d3 objects in specification."""

    pass


class Math(metaclass=JSModule):
    """JSModule class to allow Math objects in specification."""

    pass


def js(txt: str) -> dict:
    """Tag a string as JavaScript code.

    Args:
        txt (str): string containing JavaScript code.

    Returns:
        dict: tagged string as dict with type value.
    """
    return {"pyobsplot-type": "js", "value": txt}



"""
Obsplot main class.
"""

from IPython.display import display

from .widget import ObsplotWidget
from .jsdom import ObsplotJsdom
from .utils import default_renderer, available_renderers


class Obsplot:
    """
    Main Obsplot class.

    Launches a Jupyter widget with ObsplotWidget class, or displays an IPython display
    with ObsplotJsdom depending on the renderer.
    """

    # Default renderer
    renderer = default_renderer

    @staticmethod
    def set_renderer(renderer: str) -> None:
        """Set the renderer for next Obsplot calls.

        Args:
            renderer (str): renderer name.

        """
        if renderer not in available_renderers:
            raise ValueError(
                f"Incorrect renderer. Must be one of {available_renderers}."
            )
        Obsplot.renderer = renderer

    def __new__(cls, *args, **kwargs) -> None:
        """
        Constructor. Extract spec from args and kwargs, and call the configured renderer.
        """

        # Only one dict arg -> spec passed as dict
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], dict):
            spec = args[0]
        # Only one kwarg called spec
        elif len(args) == 0 and len(kwargs) == 1 and "spec" in kwargs:
            spec = kwargs["spec"]
        # Only kwargs -> spec is kwargs
        elif len(args) == 0 and len(kwargs) > 0:
            spec = kwargs
        else:
            raise ValueError("Incorrect ObsPlot arguments")

        # Plot spec with the configured renderer
        if Obsplot.renderer == "widget":
            return ObsplotWidget(spec)
        elif Obsplot.renderer == "jsdom":
            display(ObsplotJsdom(spec).plot())
        else:
            raise ValueError("Incorrect renderer.")



"""
Utility methods and variables 
"""

import pathlib

# Output directory of esbuild
bundler_output_dir = pathlib.Path(__file__).parent / "static"
# Default renderer
default_renderer = "widget"
available_renderers = ["widget", "jsdom"]



"""
Functions for DataFrame objects conversion to Arrow IPC bytes.
"""


import io
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.feather as pf

from typing import Any


def pd_to_arrow(df: pd.DataFrame) -> bytes:
    """Convert a pandas DataFrame to Arrow IPC bytes.

    Args:
        df (pd.DataFrame): pandas DataFrame to convert.

    Returns:
        bytes: Arrow IPC bytes
    """
    f = io.BytesIO()
    df.to_feather(f, compression="uncompressed")
    return f.getvalue()


def pl_to_arrow(df: pl.DataFrame) -> bytes:
    """Convert a polars DataFrame to Arrow IPC bytes.

    Args:
        df (pl.DataFrame): polars DataFrame to convert.

    Returns:
        bytes: Arrow IPC bytes.
    """

    pa_table = df.to_arrow()
    pa_table_no_big = pa_table.cast(arrow_schema_no_big(pa_table.schema))

    f = io.BytesIO()
    pf.write_feather(pa_table_no_big, f, compression="uncompressed")
    return f.getvalue()


def arrow_schema_no_big(pa_schema: Any) -> Any:
    """Transform a pyarrow schema by converting large strings to strings,
    float64 to float32 and int64 to int32. This is needed because JavaScript
    apache-arrow library doesn't support large types yet.

    Args:
        pa_schema (Any): schema to convert.

    Returns:
        Any: conversion result.
    """
    pa_schema_no_big = []
    for col_name, pa_dtype in zip(pa_schema.names, pa_schema.types):
        # Large strings to string
        if pa_dtype == pa.large_string():
            pa_schema_no_big.append(pa.field(col_name, pa.string()))
        # Int64 to Int32
        elif pa_dtype == pa.int64():
            pa_schema_no_big.append(pa.field(col_name, pa.int32()))
        # Float64 to Float32
        elif pa_dtype == pa.float64():
            pa_schema_no_big.append(pa.field(col_name, pa.float32()))
        # Large strings in dictionaries to string (for polars Categorical)
        elif pa_dtype == pa.dictionary(pa.uint32(), pa.large_string(), ordered=False):
            pa_schema_no_big.append(
                pa.field(
                    col_name, pa.dictionary(pa.uint32(), pa.string(), ordered=False)
                )
            )
        elif pa_dtype == pa.dictionary(pa.uint32(), pa.large_string(), ordered=True):
            pa_schema_no_big.append(
                pa.field(
                    col_name, pa.dictionary(pa.uint32(), pa.string(), ordered=True)
                )
            )
        else:
            pa_schema_no_big.append(pa.field(col_name, pa_dtype))
    return pa.schema(pa_schema_no_big)



"""
Obsplot widget handling.
"""


import anywidget
import traitlets

from .utils import bundler_output_dir
from .parsing import SpecParser


class ObsplotWidget(anywidget.AnyWidget):
    """Obsplot widget class.

    It inherits from anywidget.Anywidget.

    The class takes a plot specification as input and generates a plot.

    The specification can be given as a dict, a Plot function call or as
    Python kwargs.
    """

    _esm = bundler_output_dir / "widget.js"
    _css = bundler_output_dir / "styles.css"
    # spec traitlet : plot specification
    spec = traitlets.Dict().tag(sync=True)

    def __init__(self, spec):
        """Obsplot widget constructor."""

        # Init widget
        super().__init__(spec=spec)

    @traitlets.validate("spec")
    def _validate_spec(self, proposal):
        spec = proposal["value"]
        parser = SpecParser("widget")
        code = parser.parse(spec)
        spec = {"data": parser.serialize_data(), "code": code}
        return spec



# Get __version__ from pyproject.toml (currently installed)
try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version("anywidget")
except PackageNotFoundError:
    __version__ = "uninstalled"


from .obsplot import Obsplot  # noqa:F401
from .parsing import Plot, d3, Math, js  # noqa:F401

__all__ = ["Obsplot", "Plot", "d3", "Math", "js", "__version__"]
